set rcsId {$Id: widgetip.tcl,v 2.3 1999/12/27 16:24:23 jfontain Exp $}

class widgetTip {

    class topLabel {

        proc topLabel {this parentPath args} composite {
            [new toplevel $parentPath -highlightbackground black -highlightthickness 1] $args
        } {
            composite::manage $this [new label $widget::($this,path)] label
            composite::complete $this
            pack $composite::($this,label,path)
            wm overrideredirect $widget::($this,path) 1                                             ;# no window manager decorations
        }

        proc ~topLabel {this} {}

        proc options {this} {
            return [list\
                [list -bordercolor borderColor BorderColor Black Black]\
                [list -borderwidth borderWidth BorderWidth 1 1]\
                [list\
                    -background background Background\
                    $widget::(default,ButtonBackgroundColor) $widget::(default,ButtonBackgroundColor)\
                ]\
                [list -font font Font $widget::(default,ButtonFont) $widget::(default,ButtonFont)]\
                [list -foreground foreground Foreground $widget::(default,ButtonForeground) $widget::(default,ButtonForeground)]\
                [list -text text Text {} {}]\
            ]
        }

        foreach option {-background -font -foreground -text} {
            proc set$option {this value} "\$composite::(\$this,label,path) configure $option \$value"
        }

        proc set-bordercolor {this value} {
            $widget::($this,path) configure -highlightbackground $value
        }

        proc set-borderwidth {this value} {
            $widget::($this,path) configure -highlightthickness $value
        }
    }

    if {![info exists (label)]} {
        set (label) [new topLabel . -font $widget::(default,EntryFont) -background #FFFFBF]
        set (path) $widget::($(label),path)
        wm withdraw $(path)
        # handle button and key presses as global events for some child widgets (such as entries) do not pass them to their parent
        bind all <ButtonPress> {widgetTip::globalEvent %W}
        bind all <KeyPress> {widgetTip::globalEvent %W}
        set (xLast) -1
        set (yLast) -1
    }

    proc widgetTip {this args} switched {$args} {
        switched::complete $this
    }

    proc ~widgetTip {this} {
        disable $this
        if {[info exists ($this,bindings)]} {                                                          ;# eventually remove bindings
            delete $($this,bindings)
        }
    }

    proc options {this} {
        return [list\
            [list -font $widget::(default,EntryFont) $widget::(default,EntryFont)]\
            [list -path {} {}]\
            [list -text {} {}]\
        ]
    }

    proc set-path {this value} {
        if {$switched::($this,complete)} {
            error {option -path cannot be set dynamically}
        }
        if {![winfo exists $value]} {
            error "invalid widget: \"$value\""
        }
        set bindings [new bindings $value 0]
        bindings::set $bindings <Enter> "widgetTip::enable $this"
        bindings::set $bindings <Leave> "widgetTip::disable $this"
        set ($this,bindings) $bindings
    }

    proc set-font {this value} {}                                                  ;# nothing to do, data is saved at switched level
    proc set-text {this value} {}

    proc globalEvent {widget} {
        if {![catch {string first $switched::($(active),-path) $widget} value]&&($value==0)} {
            disable $(active)                        ;# hide if active widget exists and is a descendant of the active target widget
        }
    }

    proc show {this x y} {
        set path $(path)
        widget::configure $(label) -font $switched::($this,-font) -text $switched::($this,-text)                 ;# update tip label
        showTopLevel $path +$x+$y
        update idletasks
        raise $path
    }

    proc enable {this} {
        set x [winfo pointerx $(path)]
        set y [winfo pointery $(path)]
        if {($x==$(xLast))&&($y==$(yLast))} {
            widgetTip::show $this [expr {$x+7}] [expr {$y+10}]
        } else {
            set (xLast) $x
            set (yLast) $y
            set (event) [after 300 "widgetTip::enable $this"]                                                                ;# poll
        }
        set (active) $this                                                                                 ;# remember active object
    }

    proc disable {this} {
        # event and active tip may no longer exist when the pointer leaves after a click (for example)
        catch {after cancel $(event)}
        catch {unset (active)}
        wm withdraw $(path)
    }

}
